import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.*;

import sv.kernel.*;
import sv.ui.*;
import sv.util.*;
import sv.drawer.*;

/** 
 * GCPlane.java
 * Define a canvas to draw the graph
 * @version 1.0    10/15/97
 * @author Byeongseob Ki
 */

abstract class GCPlane extends GCanvas {

  // size of the canvas
  protected Dimension canvasSize;

  // canvas size of point data area 
  protected Rectangle pda;

  // axis
  protected Axis axis;

  // min and max value of given time slot
  protected float[] minmaxArr;

  // dragging
  protected boolean inversedX = false;
  protected boolean inversedY = false;
  protected int dragStartX, dragStartY, dragEndX, dragEndY;

  protected boolean isPlotAreaResize = false;
  protected int isLabelMove = -1;
  protected int selectedCorner;
  protected Rectangle pdaCorners[];

  // zooming
  protected float zoomFactorX = 1.0f, zoomFactorY = 1.0f;
  protected int zoomMode = 3;         // zoom mode 0 - xy zoom
                            //           1 - x  zoom
                            //           2 - y  zoom 
                            //           3 - no zoom 
  protected Vector zoomer;

  // to access GFPlane class
  protected GFPlane gframe;

  // to access svserver
  protected svserver svServer;
  
   // resizing
  protected float resizeFactor = 1.0f;
  protected int fontSize = 12;

  // labels
  protected RTextLine text;

  // number format
  public int xsig = 3; 
  public int ysig = 3; 
  public int timesig = 3; 
  public int xprec   = 2; 
  public int yprec   = 2; 
  public int timeprec   = 2; 
  public boolean sformat = false;


  /**
   * constructor
   */
  public GCPlane(GFPlane gframe) {
     this.gframe = gframe;
     this.svServer = gframe.svServer;

     text = new RTextLine();
     text.setColor(Color.black);
     text.setDrawingComponent(this);

     pda = new Rectangle();
     pdaCorners = new Rectangle[2];
     rect = new Rectangle();
     minmaxArr = gframe.getMinMax();
     axis = new Axis();

     makepopup();

     zoomer = new Vector();
     zoomFactorX = gframe.config.zoomFactorX;
     zoomFactorY = gframe.config.zoomFactorY;
     zoomMode = gframe.config.zoomMode;

     if (svServer.dtool != null)
        drawingmode = true;
  }  

  /**
   * return offscreen image
   */
  public Image getImage() {
     return offScrImage;
  }

  /**
   * set data plot area
   */
  public void setPDA(Rectangle rect) {
     pda = rect;
     change_pdaCorners();
  }

  /**
   * notify the size of canvas to all attached elements
   * to recalculate the data and repaint it
   */
  public void changeNotify() {
     changeNotify(getSize());
  }

  // implementation of changeNotify
  protected abstract void changeNotify(Dimension d);



  // collaboration zooming
  public void colZooms(String[] args) {
    minmaxArr[0] = Float.valueOf(args[0]).floatValue();
    minmaxArr[1] = Float.valueOf(args[1]).floatValue();
    minmaxArr[2] = Float.valueOf(args[2]).floatValue();
    minmaxArr[3] = Float.valueOf(args[3]).floatValue();

    zoomFactorX = Float.valueOf(args[4]).floatValue();
    zoomFactorY = Float.valueOf(args[5]).floatValue();
    zoomMode = Integer.valueOf(args[6]).intValue();

    gframe.config.zoomStartX = minmaxArr[0];
    gframe.config.zoomStartY = minmaxArr[1];
    gframe.config.zoomEndX = minmaxArr[2];
    gframe.config.zoomEndY = minmaxArr[3];
    gframe.config.zoomFactorX = zoomFactorX;
    gframe.config.zoomFactorY = zoomFactorY;
    gframe.config.zoomMode = zoomMode;

    zoomer.addElement(new ZoomHandle(minmaxArr[0], minmaxArr[1],
         minmaxArr[2], minmaxArr[3], zoomFactorX, zoomFactorY, zoomMode));

    repaint();
  }


  public void resetZooming() {
     while (!zoomer.isEmpty()) {
        ZoomHandle zh = (ZoomHandle)zoomer.lastElement();
        zoomer.removeElement(zh);
        zh = null;
     }
     gframe.zoomoutMI.setEnabled(false);
     zoomFactorX = 1.0f;
     zoomFactorY = 1.0f;
     minmaxArr = gframe.getMinMax();
     repaint();
  }

  public void zoomOut() {
    ZoomHandle zh;

    if (!zoomer.isEmpty()) {
        zh = (ZoomHandle)zoomer.lastElement();
        zoomer.removeElement(zh);

    } else {
        gframe.zoomoutMI.setEnabled(false);
        return;
    }

    String args[] = zh.getAllInfo();

    minmaxArr[0] = Float.valueOf(args[0]).floatValue();
    minmaxArr[1] = Float.valueOf(args[1]).floatValue();
    minmaxArr[2] = Float.valueOf(args[2]).floatValue();
    minmaxArr[3] = Float.valueOf(args[3]).floatValue();

    zoomFactorX = Float.valueOf(args[4]).floatValue();
    zoomFactorY = Float.valueOf(args[5]).floatValue();
    zoomMode = Integer.valueOf(args[6]).intValue();

    gframe.config.zoomStartX = minmaxArr[0];
    gframe.config.zoomStartY = minmaxArr[1];
    gframe.config.zoomEndX = minmaxArr[2];
    gframe.config.zoomEndY = minmaxArr[3];
    gframe.config.zoomFactorX = zoomFactorX;
    gframe.config.zoomFactorY = zoomFactorY;
    gframe.config.zoomMode = zoomMode;

    if (zoomer.isEmpty()) 
       gframe.zoomoutMI.setEnabled(false);

    zh = null;
    repaint();
  }



  // draw objects
  protected void objectsDraw(Graphics g) {

     for (int i = 0; i < dh.drawingVec.size(); i++) {
        ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(i);

        if (s.getType() == ShapeProperties.IMAGE) {
           Image img = ((ImageShape)s).getImage();
           Rectangle rect = ((ImageShape)s).getBounds();
           g.drawImage(img, rect.x, rect.y, rect.width, rect.height, this);
        } else {
           s.setBGColor(gframe.config.bgColor);
           s.paint(g);
        }
       
        if (drawingmode && s.isFocused()) {
           Color saved = g.getColor();
           
           Vector vecShapes = s.getHotSpots();
           if (vecShapes == null) return;

           g.setColor(Color.lightGray);
           for (int j = 0; j < vecShapes.size(); j++) {
             HotSpot hs = (HotSpot)vecShapes.elementAt(j);
             hs.paint(g);
           }
           g.setColor(saved);
        }
     }
  }

  // find PDA coners
  protected void change_pdaCorners() {
     pdaCorners[0] = new Rectangle(pda.x-10, pda.y-10, 20, 20);
     pdaCorners[1] = 
         new Rectangle(pda.x+pda.width-10, pda.y+pda.height-10, 20, 20);
  }

  // Find the closest data point to the cursor
  protected float[] getClosestPoint(float x, float y) {
     float a[] = new float[3];
     float distsq = -1.0f;
     float data[] = {0.0f, 0.0f};

     for (int i=0; i<gframe.dataSetVector.size(); i++) {
        DSPlane ds = (DSPlane)gframe.dataSetVector.elementAt(i);

        for (int j=0; j<ds.getTimeArr().length; j++) {
           if (ds.getTimeValue(j) == gframe.timeArr[gframe.currentTime]) {

               a = ds.timeData[j].getClosestPoint(x,y);

               if(distsq < 0.0f || distsq > a[2]) {
                    data[0] = a[0];
                    data[1] = a[1];
                    distsq  = a[2];
	        }
	    }
         }
     }
     return data;
  }

  /**
   * do whiteboard related work
   */
  public void doWhiteboardWork(int id) {
     int sindex = -1;
     ShapeObject sobject;

     for (int i = 0; i < dh.drawingVec.size(); i++) {
        ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(i);

        if (s.isFocused()) {
           sindex = i;
           sobject = s;
           break;
        }
     }

     String args[] = new String[1];
     switch (id) {
        case CommandSet.WCUT:
           if (sindex < 0) return;
           args = new String[1];
           args[0] = String.valueOf(sindex);
           svServer.fmanager.action1d(gframe, id, args, true);           
           break;

        case CommandSet.WCOPY:
           if (sindex < 0) return;
           args = new String[1];
           args[0] = String.valueOf(sindex);
           svServer.fmanager.action1d(gframe, id, args, true);  
           break;

        case CommandSet.WPASTE:
           svServer.fmanager.action1d(gframe, id, null, true);  
           break;

        case CommandSet.WFILL:
           if (sindex < 0) return;
           args = new String[1];
           args[0] = String.valueOf(sindex);
           svServer.fmanager.action1d(gframe, id, args, true);
           break;

        case CommandSet.WLINECOLOR:

           args = new String[2];
           args[0] = String.valueOf(sindex);
           args[1] = String.valueOf(svServer.shapes.getLineColor().getRGB());
           svServer.fmanager.action1d(gframe, id, args, true);
           break;

        case CommandSet.WFILLCOLOR:
           args = new String[2];
           args[0] = String.valueOf(sindex);
           args[1] = String.valueOf(svServer.shapes.getFillColor().getRGB());
           svServer.fmanager.action1d(gframe, id, args, true);
           break;

        case CommandSet.WTEXTCOLOR:
           args = new String[2];
           args[0] = String.valueOf(sindex);
           args[1] = String.valueOf(svServer.shapes.getTextColor().getRGB());
           svServer.fmanager.action1d(gframe, id, args, true);
           break;

        case CommandSet.WTEXTFONT:
           args = new String[4];
           Font f = svServer.shapes.getTextFont();
           args[0] = String.valueOf(sindex);
           args[1] = f.getName();
           args[2] = String.valueOf(f.getStyle());
           args[3] = String.valueOf(f.getSize());
           svServer.fmanager.action1d(gframe, id, args, true);
           break;

        case CommandSet.WPEN:
           args = new String[4];

           args[0] = String.valueOf(sindex);
           args[1] = String.valueOf(svServer.shapes.getLineType());
           args[2] = String.valueOf(svServer.shapes.getLineWidth());
           args[3] = String.valueOf(svServer.shapes.getDashSize());
           svServer.fmanager.action1d(gframe, id, args, true);
           break;
     }
  }

  /**
   * do add point action
   */
  public void doWAddPoint(int type, int x, int y) {

     ShapeObject s = (ShapeObject)dh.drawingVec.lastElement();
     
     if (s.getType() != type) {
       multiShape = null;
       svServer.dtool.clearAllToggleButtons();  
     }

     if (type == ShapeProperties.MULTILINE) {
        HotSpot hs = ((MLineShape)s).addPoint(new Point(x, y));
        dh.hotspotManager.manage(hs);
     } else if (type == ShapeProperties.BEZIER) {
        HotSpot hs = ((BLineShape)s).addPoint(new Point(x, y));
        dh.hotspotManager.manage(hs);
     }

     repaint();
  }


  /**
   * do add points action
   */
  public void doWAddPoints(String[] args) {

     ShapeObject s = (ShapeObject)dh.drawingVec.lastElement();
     
     if (s.getType() != Integer.valueOf(args[0]).intValue()) {
       multiShape = null;
       svServer.dtool.clearAllToggleButtons();  
     }

     int sz = Integer.valueOf(args[1]).intValue();
     Point[] pts = new Point[sz];

     int idx = 2;
     for (int i=0; i<sz; i++) {
        pts[i] = new Point(Integer.valueOf(args[idx]).intValue(),
                           Integer.valueOf(args[idx+1]).intValue());
        idx = idx+2;
     }

     if (s.getType() == ShapeProperties.MOUSE) 
        ((RLineShape)s).addPoints(pts);

     else if (s.getType() == ShapeProperties.POLYGON) {
        HotSpot[] hs = ((PolygonShape)s).addPoints(pts);
        if (hs == null) return;

        for (int i=0; i<hs.length; i++)
           dh.hotspotManager.manage(hs[i]);
     }
     repaint();
  }



  // make popup menu for the whiteboard
  protected void makepopup() {

     wpopup = new PopupMenu("Drawing tools:");

     wcutmi = new MenuItem("Cut");
     wcutmi.addActionListener(new CSPlane(gframe, CommandSet.WCUT));
     wpopup.add(wcutmi);

     wcopymi = new MenuItem("Copy");
     wcopymi.addActionListener(new CSPlane(gframe, CommandSet.WCOPY));
     wpopup.add(wcopymi);

     wpastemi = new MenuItem("Paste");
     wpastemi.addActionListener(new CSPlane(gframe, CommandSet.WPASTE));
     wpopup.add(wpastemi);

     wpopup.addSeparator();

     wfillmi = new MenuItem("Fill");
     wfillmi.addActionListener(new CSPlane(gframe, CommandSet.WFILL));
     wpopup.add(wfillmi);

     wlinecmi = new MenuItem("Apply line color");
     wlinecmi.addActionListener(new CSPlane(gframe, CommandSet.WLINECOLOR));
     wpopup.add(wlinecmi);

     wlinepmi = new MenuItem("Apply line properties");
     wlinepmi.addActionListener(new CSPlane(gframe, CommandSet.WPEN));
     wpopup.add(wlinepmi);

     wfillcmi = new MenuItem("Apply fill color");
     wfillcmi.addActionListener(new CSPlane(gframe, CommandSet.WFILLCOLOR));
     wpopup.add(wfillcmi);

     wtextcmi = new MenuItem("Apply text color");
     wtextcmi.addActionListener(new CSPlane(gframe, CommandSet.WTEXTCOLOR));
     wpopup.add(wtextcmi);

     wtextfmi = new MenuItem("Apply text font");
     wtextfmi.addActionListener(new CSPlane(gframe, CommandSet.WTEXTFONT));
     wpopup.add(wtextfmi);

     add(wpopup); // add popup menu
  }

} // end of class
